/*
  Macro           FindNoScroll
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   Probably all versions, all variants
  Version         v1   25 Feb 2024

  This tool works like TSE's Find command, but with no scrolling if the found
  occurrence was already on the screen.


  INSTALLATION

  Copy this file to TSE's "mac" directory and compile it there, for example by
  opening the file there in TSE, and applying the Macro Compile menu.


  USAGE

  Either execute the tool from TSE's Macro Execute menu,
  or from the Potpourri menu,
  or assign keys to:
    ExecMacro("FindNoScroll")
    ExecMacro("FindNoScroll repeat")


  KNOWN IMPERFECTION

  The tool works less pretty if the screen is or becomes horizontally scrolled.


  HISTORY

  v1   25 Feb 2024
    Initial version.

*/


// Global constants
#define S_O_LENGTH 11   // Length of search options field.


proc optimize_search_options(var string search_options)
  integer p = 0

  // Remove meaningless characters
  p = 1
  while p <= Length(search_options)
    if not Pos(Lower(search_options[p]), 'abcgilvwx^$+')
      search_options = search_options[1: p - 1] + search_options[p + 1: S_O_LENGTH]
    else
      p = p + 1
    endif
  endwhile

  // Remove duplicate characters
  p = 1
  while p < Length(search_options)
    if Pos(Lower(search_options[p]), Lower(search_options[p + 1: S_O_LENGTH]))
      search_options = search_options[1: p - 1] + search_options[p + 1: S_O_LENGTH]
    else
      p = p + 1
    endif
  endwhile
end optimize_search_options


proc remove_search_options(var string search_options, string removables)
  integer p = 0
  integer r = 0

  for r = 1 to Length(removables)
    p = Pos(Lower(removables[r]), Lower(search_options))
    while p
      search_options = search_options[1: p - 1] + search_options[p + 1: S_O_LENGTH]
      p = Pos(Lower(removables[r]), Lower(search_options))
    endwhile
  endfor
end remove_search_options


proc common_find(string search_string, string search_options_in)
  integer fnd_currline                 = 0
  integer fnd_displayed_line_from      = 0
  integer fnd_row                      = 0
  string  fnd_text      [MAXSTRINGLEN] = ''
  integer old_currline                 = 0
  integer old_displayed_line_from      = 0
  integer old_displayed_line_to        = 0
  string  search_options  [S_O_LENGTH] = search_options_in

  old_currline            = CurrLine()
  old_displayed_line_from = old_currline - CurrRow() + 1
  old_displayed_line_to   = old_displayed_line_from + Query(WindowRows) - 1

  UpdateDisplay(_ALL_WINDOWS_REFRESH_)  // Undo any old find-hiliting.

  if lFind(search_string, search_options)
    if Pos('v', Lower(search_options))
      // Make room to add "cg" options.
      optimize_search_options(search_options)
      remove_search_options(search_options, 'cgv')
      search_options = 'cg' + search_options

      // Now refind the string, but in the selected line.
      lFind(search_string, search_options)
    endif

    fnd_currline            = CurrLine()
    fnd_row                 = CurrRow()
    fnd_text                = GetFoundText()
    fnd_displayed_line_from = fnd_currline - CurrRow() + 1

    if (fnd_currline in old_displayed_line_from .. old_displayed_line_to)
      if     fnd_displayed_line_from > old_displayed_line_from
        ScrollUp(fnd_displayed_line_from - old_displayed_line_from)
        fnd_row = fnd_row + (fnd_displayed_line_from - old_displayed_line_from)
      elseif fnd_displayed_line_from < old_displayed_line_from
        ScrollDown(old_displayed_line_from - fnd_displayed_line_from)
        fnd_row = fnd_row - (old_displayed_line_from - fnd_displayed_line_from)
      endif
    endif

    // First let TSE update its own hiliting.
    UpdateDisplay(_ALL_WINDOWS_REFRESH_)

    PutStrAttrXY(Query(WindowX1) + CurrCol() - 1,
                 Query(WindowY1) + fnd_row - 1,
                 fnd_text,
                 '',
                 Query(HiliteAttr))
  else
    if Query(Beep)
      Alarm()
    endif
    Message(search_string; 'not found.')
  endif
end common_find


proc repeat_find()
  string search_options   [S_O_LENGTH] = ''
  string search_string  [MAXSTRINGLEN] = ''

  search_string  = GetHistoryStr(_FIND_HISTORY_       , 1)
  search_options = GetHistoryStr(_FINDOPTIONS_HISTORY_, 1)

  optimize_search_options(search_options)
  remove_search_options(search_options, 'g')
  search_options = '+' + search_options

  common_find(search_string, search_options)
end repeat_find


proc new_find()
  string  search_options  [S_O_LENGTH] = ''
  string  search_string [MAXSTRINGLEN] = ''

  if  Ask('Search for:                              (when possible without scrolling)',
          search_string, _FIND_HISTORY_)
  and Ask('Options [ABGLIWX] (All-files Back Global Local Ignore-case Words reg-eXp):',
          search_options, _FINDOPTIONS_HISTORY_)
    common_find(search_string, search_options)
  endif
end new_find


proc Main()
  string cmd [MAXSTRINGLEN] = ''

  cmd = Query(MacroCmdLine)

  if Pos('repeat', Lower(cmd))
    repeat_find()
  else
    new_find()
  endif

  PurgeMacro(CurrMacroFilename())
end Main


